import { addComponent, addEntity } from '../ecs.js'
import { PrioritySchedule, SystemSchedule, Schedules } from './schedule.js'

/** @typedef {(engine: Engine) => void} Plugin */
/** @typedef {number} Entity */

export class Startup { }
export class Update { }
export class FixedUpdate { }
export class Render { }

export class Time {
    /** @type {DOMHighResTimeStamp} */
    currentTime

    /** @type {DOMHighResTimeStamp} */
    deltaTime
}

class Resources {
    #inner = new Map()

    /**
     * @template T
     * @param {T} resource
     */
    add(resource) {
        this.#inner.set(resource.constructor, resource)
    }

    /**
     * @template T
     * @param {new(...args: any[]) => T} resource
     * @returns {T}
     */
    get(resource) {
        return this.#inner.get(resource)
    }
}


export class World {
    #resources = new Resources()

    /**
     * @template T
     * @param {T} resource
     */
    addResource(resource) {
        this.#resources.add(resource)
    }

    /**
     * @template T
     * @param {new(...args: any[]) => T} resource
     * @returns {T}
     */
    getResource(resource) {
        return this.#resources.get(resource)
    }
}

export class Canvas {
    /** @type {HTMLCanvasElement} */
    #element

    /**
     * @param {HTMLCanvasElement} element
     */
    constructor(element) {
        this.#element = element
    }

    get element() {
        return this.#element
    }
}

// TODO: add analog to Commands/EntityCommands
export class Engine {
    /** @type {World} */
    #world

    /** @type {Plugin[]} */
    #plugins = []

    #schedules = new Schedules()

    /** @type {number} */
    #loopId

    #initialized = false

    /**
    * @param {World} world
    * @param {HTMLCanvasElement} canvas
    */
    constructor(world, canvas) {
        this.#world = world

        this.addResource(new Time())
        this.addResource(new Canvas(canvas))

        this.#schedules.add(
            Startup,
            new PrioritySchedule(0, new SystemSchedule())
        )

        this.#schedules.add(
            Update,
            new PrioritySchedule(1, new SystemSchedule())
        )

        this.#schedules.addBefore(
            Update,
            FixedUpdate,
            new SystemSchedule()
        )

        this.#schedules.addAfter(
            Update,
            Render,
            new SystemSchedule()
        )
    }

    get world() {
        return this.#world
    }

    /**
     * @param {any} resource 
     * @returns {Engine}
     */
    addResource(resource) {
        this.#world.addResource(resource)
        return this
    }

    /**
     * @template T
     * @param {new(...args: any[]) => T} resource
     * @returns {T}
     */
    getResource(resource) {
        return this.#world.getResource(resource)
    }

    /**
     * @param {import('./schedule.js').ScheduleLabel} schedule 
     * @param {import('./schedule').System} system 
     * @param {import('./schedule').RegisterSystemOptions} [options={}]
     * @returns {Engine}
     */
    addSystem(schedule, system, options) {
        this.#schedules.get(schedule).register(system, options)
        return this
    }

    /**
    * @param {Plugin | Plugin[]} plugin
    * @returns {Engine}
    */
    addPlugin(plugin) {
        if (Array.isArray(plugin)) {
            plugin.forEach(this.addPlugin.bind(this))
        } else {
            plugin(this)
        }

        return this
    }

    stop() {
        cancelAnimationFrame(this.#loopId)
        this.#loopId = undefined
    }

    async run() {
        // FIXME: modularize this
        const startup = this.#schedules.get(Startup).schedule
        const main = this.#schedules.get(Update).schedule
        const fixed = this.#schedules.get(FixedUpdate).schedule
        const render = this.#schedules.get(Render).schedule

        const fixedTimeStep = 1000 / 60

        let lastTime = performance.now()
        let accumulator = 0

        /** @type (time: DOMHighResTimeStamp) => void */
        const loop = currentTime => {
            const deltaTime = currentTime - lastTime
            lastTime = currentTime
            accumulator += deltaTime

            const time = this.getResource(Time)

            time.currentTime = currentTime
            time.deltaTime = deltaTime

            while (accumulator >= fixedTimeStep) {
                fixed.execute(this.#world)
                accumulator -= fixedTimeStep
            }

            main.execute(this.#world)
            render.execute(this.#world)

            this.#loopId = requestAnimationFrame(loop)
        }

        await startup.execute(this.#world)
        this.#loopId = requestAnimationFrame(loop)
    }
}

