import { PriorityMap } from '../utils/ordered-map.js'

/** @typedef {Function} ScheduleLabel */

/**
 * @template Dependency
 */
export class CyclicalDependencyError extends Error {
    /**
     * @param {Dependency} dependency
     */
    constructor(dependency) {
        super(`Cyclical dependency: ${dependency}`)
    }
}

/**
 * @template Dependency
 */
export class UnknownDependencyError extends Error {
    /**
     * @param {Dependency} dependency
     */
    constructor(dependency) {
        super(`Dependency ${dependency} not found`)
    }
}


/**
 * @template Node
 */
export class Visitor {
    #nodes
    #recursion
    #visited

    /**
    * Create a Visitor
    * @param {Iterable.<Node>} nodes
    */
    constructor(nodes) {
        this.#nodes = Array.from(nodes)
        this.#visited = new Set()
        this.#recursion = new Set()
    }

    /** @typedef {(node: Node) => void} VisitNode */
    /** @typedef {(node: Node, visit: VisitNode) => void} HandleNode */

    /**
     * @param {Node} node
     * @param {HandleNode} onVisit
     *
     * @throws {CircularDependencyError}
     * Thrown if two nodes rely on one another
     */
    _visitNode(node, onVisit) {
        if (this.#recursion.has(node)) {
            throw new CyclicalDependencyError(node)
        }

        if (this.#visited.has(node)) return

        this.#visited.add(node)
        this.#recursion.add(node)

        onVisit(node, node => this._visitNode(node, onVisit))

        this.#recursion.delete(node)
    }

    /**
    * Visit all nodes with the onVisit callback
    *
    * @param {HandleNode} onVisit
    * @returns {Node[]}
    */
    visit(onVisit) {
        const len = this.#nodes.length

        for (let i = 0; i < len; i++) {
            this._visitNode(this.#nodes[i], onVisit)
        }

        return Array.from(this.#visited.values())
    }

    /**
    * Reset the visitor
    */
    clear() {
        this.#visited.clear()
        this.#recursion.clear()
    }
}

export class Schedule {
    execute(world) {
        throw new NotImplementedError()
    }
}

/** @typedef {(world: any) => void} System */
/**
 * @typedef {Object} SystemSchema
 * @property {System} system
 * @property {Set<System>} dependencies
 * @property {Set<System>} before
 * @property {Set<System>} after
 */

/**
 *
 * @typedef {Object} RegisterSystemOptions
 * @property {System[]} [options.dependencies = []]
 * @property {System} [options.before]
 * @property {System} [options.after]
 */

export class SystemSchedule {
    /** @type {Map<System, SystemSchema>} */
    _graph = new Map()

    /** @type {System[]?} */
    _cache = undefined

    /**
     * @param {System} node
     * @returns {SystemSchema}
     *
     * @throws {UnknownDependencyError}
     * Thrown if the requested node is not in the dependency graph
     */
    _tryGet(node) {
        if (!this._graph.has(node)) {
            throw new UnknownDependencyError(node)
        }

        return this._graph.get(node)
    }

    /**
     * @param {SystemSchema} beforeNode
     * @param {SystemSchema} afterNode
     */
    _insert(beforeNode, afterNode) {
        afterNode.before.add(beforeNode.system)
        beforeNode.after.add(afterNode.system)
    }

    /**
    * Register a new system with this schedule
    *
    * @param {System} system
    * @param {RegisterSystemOptions} [options={}]
    */
    register(system, { dependencies = [], before, after } = {}) {
        this._cache = undefined

        this._graph.set(system, { system, dependencies: new Set(dependencies), before: new Set(), after: new Set() })

        const systemNode = this._graph.get(system)

        for (const dep of dependencies) {
            const depNode = this._tryGet(dep)
            depNode.after.add(system)
        }

        if (before) {
            const beforeNode = this._tryGet(before)
            this._insert(systemNode, beforeNode)
        }

        if (after) {
            const afterNode = this._tryGet(after)
            this._insert(afterNode, systemNode)
        }
    }

    /**
    * @returns {System[]}
    * 
    * @throws {UnknownDependencyError}
    * Thrown if a dependency was not registered before resolution
    *
    * @throws {CircularDependencyError}
    * Thrown if two nodes rely on one another
    */
    resolve() {
        if (this._cache) {
            return this._cache
        }

        const visitor = new Visitor(this._graph.keys())

        this._cache = visitor.visit(
            (next, visit) => {
                const node = this._graph.get(next)

                for (const dep of node.dependencies) {
                    visit(dep)
                }

                for (const before of node.before) {
                    visit(before)
                }
            }
        )


        return this._cache
    }

    /**
    * @param {any} world
    *
    * @throws {UnknownDependencyError}
    * Thrown if a dependency was not registered before resolution
    *
    * @throws {CircularDependencyError}
    * Thrown if two nodes rely on one another
    */
    async execute(world) {
        const systems = this.resolve()
        const len = systems.length
        for (let i = 0; i < len; i++) {
            await systems[i](world)
        }
    }
}

export class PrioritySchedule {
    /** @type {Number} */
    priority

    /** @type {SystemSchedule} */
    schedule

    /**
     * @param {Number} priority
     * @param {SystemSchedule} schedule
     */
    constructor(priority, schedule) {
        this.priority = priority
        this.schedule = schedule
    }

    register(system, options) {
        this.schedule.register(system, options)
    }

    async execute(world) {
        await this.schedule.execute(world)
    }
}

/**
 * @param {[ScheduleLabel, PrioritySchedule]} value
 * @return {number}
 */
const getPriority = value => value[1].priority

/**
 * @extends PriorityMap<ScheduleLabel, PrioritySchedule>
 */
export class Schedules extends PriorityMap {
    /** @type {Number} */
    #lowest

    /** @type {Number} */
    #highest

    constructor() {
        super(getPriority)

        this.#lowest = -Infinity
        this.#highest = Infinity
    }

    /**
     * @param {ScheduleLabel} target
     * @param {ScheduleLabel} label
     * @param {SystemSchedule} schedule
     */
    addBefore(target, label, schedule) {
        const targetSchedule = this.get(target)
        let priority = this.#lowest

        if (targetSchedule) {
            priority = targetSchedule.priority - 1
        }

        this.add(label, new PrioritySchedule(priority, schedule))
    }

    /**
     * @param {ScheduleLabel} target
     * @param {ScheduleLabel} label
     * @param {SystemSchedule} schedule
     */
    addAfter(target, label, schedule) {
        const targetSchedule = this.get(target)
        let priority = this.#lowest

        if (targetSchedule) {
            priority = targetSchedule.priority + 1
        }

        this.add(label, new PrioritySchedule(priority, schedule))
    }

    /**
     * @param {ScheduleLabel} label
     * @param {PrioritySchedule} prioritySchedule
     */
    add(label, prioritySchedule) {
        const priority = prioritySchedule.priority

        if (priority > this.#lowest) {
            this.#lowest = priority
        } else if (priority < this.#highest) {
            this.#highest = priority
        }

        this.set(label, prioritySchedule)
    }
}

