import { VertexFormat } from './vertex-format.js'

/**
 * @typedef {number} MeshVertexAttributeId
 * @typedef {number} BufferAddress
 */

export class VertexAttribute {
    #format
    #offset
    #shaderLocation

    /**
     * @param {VertexFormat} format 
     * @param {number} offset 
     * @param {number} shaderLocation 
     */
    constructor(format, offset, shaderLocation) {
        this.#format = format
        this.#offset = offset
        this.#shaderLocation = shaderLocation
    }

    get format() { return this.#format }
    get offset() { return this.#offset }
    get shaderLocation() { return this.#shaderLocation }
}

export class MeshVertexAttribute {
    #name
    #id
    #format

    /**
     * @param {string} name 
     * @param {MeshVertexAttributeId} id 
     * @param {VertexFormat} format 
     */
    constructor(name, id, format) {
        this.#name = name
        this.#id = id
        this.#format = format
    }

    get name() { return this.#name }
    get id() { return this.#id }
    get format() { return this.#format }
}

export class VertexAttributeDescriptor {
    #id
    #name
    #shaderLocation

    /**
     * @param {MeshVertexAttributeId} id 
     * @param {string} name 
     * @param {number} location 
     */
    constructor(id, name, location) {
        this.#id = id
        this.#name = name
        this.#shaderLocation = location
    }

    get id() { return this.#id }
    get name() { return this.#name }
    get shaderLocation() { return this.#shaderLocation }
}

class VertexAttributeError extends Error {
    /**
     * @param {MeshVertexAttributeId} id 
     * @param {string} name 
     * @param {string} message 
     */
    constructor(id, name, message) {
        super(message)
        this.id = id
        this.name = name
    }

    /**
     * @param {MeshVertexAttributeId} id 
     * @param {string} name 
     */
    static missing(id, name) {
        return new VertexAttributeError(id, name, `Missing vertex attribute ${name} (id: ${id})`)
    }
}

export class VertexBufferLayout {
    #stride
    #stepMode
    #attributes

    /**
     * @param {BufferAddress} stride 
     * @param {GPUVertexStepMode} stepMode 
     * @param {VertexAttribute[]} attributes 
     */
    constructor(stride, stepMode, attributes) {
        this.#stride = stride
        this.#stepMode = stepMode
        this.#attributes = attributes
    }

    get stride() { return this.#stride }
    get stepMode() { return this.#stepMode }
    get attributes() { return this.#attributes }

    /**
     * @param {GPUVertexStepMode} mode 
     * @param {VertexFormat[]} formats
     */
    static fromVertexFormats(mode, formats) {
        let offset = 0
        const attributes = []

        const len = formats.length
        for (let i = 0; i < len; i++) {
            const format = formats[i]

            attributes.push(new VertexAttribute(
                format,
                offset,
                i
            ))

            offset += format.size
        }

        return new VertexBufferLayout(
            offset,
            mode,
            attributes
        )
    }
}

export class MeshVertexBufferLayout {
    #ids
    #layout

    /**
     * @param {Set<MeshVertexAttributeId>} ids 
     * @param {VertexBufferLayout} layout 
     */
    constructor(ids, layout) {
        this.#ids = ids
        this.#layout = layout
    }

    get attributeIds() { return this.#ids }
    get layout() { return this.#layout }

    /**
     * @param {MeshVertexAttributeId} id 
     */
    has(id) {
        return this.#ids.has(id)
    }

    /**
     * @param {VertexAttributeDescriptor[]} descriptors 
     */
    getLayout(descriptors) {
        let attributes = []

        const len = descriptors.length
        for (let i = 0; i < len; i++) {
            const desc = descriptors[i]
            const index = descriptors.indexOf(desc)

            if (index <= -1) {
                throw VertexAttributeError.missing(desc.id, desc.name)
            }

            const attr = this.#layout.attributes[index]
            attributes.push(new VertexAttribute(
                attr.format,
                attr.offset,
                desc.shaderLocation
            ))
        }

        return new VertexBufferLayout(
            this.#layout.stride,
            this.#layout.stepMode,
            attributes
        )
    }
}

// TODO: revisit this to either optimize the sparse array,
// model it more like Bevy's tagged union, or refactor
// it entirely to better suit JS
export class MeshAttributeData {
    /** @type {MeshVertexAttribute} */
    #attribute
    /** @type {number[]} */
    #values

    /**
     * @param {MeshVertexAttribute} attribute 
     * @param {number[]} values 
     */
    constructor(attribute, values) {
        this.#attribute = attribute
        this.#values = values
    }
}

