const ResizeObserverBox = Object.freeze({
    ContentBox: 'content-box',
    BorderBox: 'border-box',
    DevicePixelContentBox: 'device-pixel-content-box',
})

export class ResizeHandler {
    /** @type {HTMLCanvasElement} */
    #canvas

    /** @type {ResizeObserver} */
    #observer

    /**
     * @typedef {(width: number, height: number) => void} Subscriber
     * @type {Set<Subscriber>}
     */
    #subscribers = new Set()

    /**
     * @param {HTMLCanvasElement} canvas 
     * @param {number} dpr 
     * @param {number} limit 
     */
    constructor(canvas, dpr, limit) {
        this.#canvas = canvas

        this.#observer = new ResizeObserver(entries => {
            for (const entry of entries) {
                /** @type {HTMLCanvasElement} */
                // @ts-ignore
                const canvas = entry.target

                const size = this._getSize(entry)[0]
                let width = size.inlineSize * dpr
                let height = size.blockSize * dpr
                width = this._clamp(width, 1, limit)
                height = this._clamp(height, 1, limit)

                if (width !== canvas.width || height !== canvas.height) {
                    canvas.width = width
                    canvas.height = height
                    this.#notify(width, height)
                }
            }
        })
    }

    /**
     * @param {Subscriber} fn 
     */
    subscribe(fn) {
        this.#subscribers.add(fn)
    }

    /**
     * @param {number} width 
     * @param {number} height 
     */
    #notify(width, height) {
        this.#subscribers
            .forEach(fn => fn(width, height))
    }

    /**
     * @param {Subscriber} fn 
     */
    unsubscribe(fn) {
        this.#subscribers.delete(fn)
    }

    start() {
        this.#observer.observe(
            this.#canvas,
            { box: ResizeObserverBox.DevicePixelContentBox }
        )

    }

    stop() {
        this.#observer.unobserve(this.#canvas)
    }

    /**
     * @param {ResizeObserverEntry} entry 
     */
    _getSize(entry) {
        return entry.devicePixelContentBoxSize || entry.contentBoxSize
    }

    /**
     * @param {number} size 
     * @param {number} min 
     * @param {number} max 
     */
    _clamp(size, min, max) {
        return Math.max(min, Math.min(size, max))
    }
}
